/*
 * Minimalist Kernel Debugger - Architecture Dependent Stack Traceback
 *
 * Written May 2002 by Paul Lai <paul_lai@comcast.net>
 *
 */

#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/kallsyms.h>
#include <linux/kdb.h>
#include <linux/kdbprivate.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#define DEBUG_KDBA_BT	0


/*
 * bt_print_one
 *
 *	Print one back trace entry.
 *
 * Inputs:
 *	esp	Current stack pointer
 *	epc	Current program counter.
 *	symtab	Information about symbol that eip falls within.
 *	sp_size	size of the current stack frame
 *	argcount Maximum number of arguments to print.	-- not used
 * Outputs:
 *	None.
 * Returns:
 *	None.
 * Locking:
 *	None.
 * Remarks:
 *	None.
 */

static void
bt_print_one(kdb_machreg_t epc, kdb_machreg_t esp, const kdb_machreg_t sp_size,
	     const kdb_symtab_t *symtab, int argcount)
{
	int	i;

	if (esp)
		kdb_printf("0x%08lx", esp);
	else
		kdb_printf("          ");
	kdb_symbol_print(epc, symtab, KDB_SP_SPACEB|KDB_SP_VALUE);

	kdb_printf("(");
	for (i=0 ; i < 3 ; i++)
	{
		kdb_printf(" 0x%x,", *(unsigned int *) (esp + sp_size + (i*4)));
	}
	kdb_printf(" 0x%x", * (unsigned int *) (esp + sp_size + 16));
	kdb_printf(")");
	kdb_printf("\n");
}

#define MIPS_KERNEL_BEGIN	0x80100000
#define MOVEFPSP	0x03a0f021	/* move fp,sp	*/
#define ADDIUSPSP	0x27bd0000	/* addiu sp,sp,x	*/
#define ADDIUSPXX	0x241d0000	/* addiu sp,x,y	*/
#define ADDIUSPXX_MASK	0xFC1F0000	/* addiu sp,x,y	*/
#define SWRASP		0xafbf0000	/* sw ra,x(sp)	*/
#define JRRA		0x03e00008	/* jr ra */

#define OP_CODE(x)	((x) & ~0xffff)
#define OP_ARG(x)	((x) & 0xffff)

/*
 * Given EPC, attempt to find the stackframe associated w/ it.
 * Go backwards in the instructions (opcodes) looking for the MOVEFPSP.
 * From the next "addiu sp,sp,x" instruction, we can find the size of the
 * current stack, and the RA (if it's saved off).
 *
 */
int
kdba_find_ef( kdb_machreg_t epc, kdb_machreg_t *size, kdb_machreg_t *ra_sp)
{
	unsigned long	*lepc = (unsigned long*) epc;
	short		sp_size;

	*size = 0;
	*ra_sp = 0;

#if DEBUG_KDBA_BT
kdb_printf("kdba_find_ef( epc=0x%x)\n", (unsigned int) epc);
#endif

	for ( ; lepc
			&& ((lepc > 0x400000 && lepc < 0x7f00000) || lepc > 0x80000000)
			&& access_ok(VERIFY_READ, lepc, sizeof(unsigned int)) ;
				lepc--)
	{
		if ( (OP_CODE(*lepc)&ADDIUSPXX_MASK)  ==  ADDIUSPXX )
		{
			sp_size = OP_ARG( *lepc ) ;
			*size = sp_size;
			if (! *ra_sp == 0)
			{
#if DEBUG_KDBA_BT
kdb_printf("kdba_find_ef( ): return *size=%d\n", (int) *size);
#endif
				return *size;
#if DEBUG_KDBA_BT
			} else {
				kdb_printf("kdba_find_ef(): addiu sp,x,y found before sw ra,sp\n");
#endif
			}
		}
		if ( OP_CODE(*lepc) == SWRASP )
		{
			*ra_sp = OP_ARG( *lepc ) ;
#if DEBUG_KDBA_BT
kdb_printf("kdba_find_ef( ): ra_sp=%d\n", (int) *ra_sp);
#endif
		}
	}

#if DEBUG_KDBA_BT
	kdb_printf("kdba_find_ef(): why did we stop here? lepc=0x%x\n", lepc);
#endif
	return 0;
}

/*
 * kdba_bt_stack_mips
 *
 *	kdba_bt_stack with mips specific parameters.
 *	Specification as kdba_bt_stack plus :-
 *
 * Inputs:
 *	As kba_bt_stack plus
 *	regs_esp If 1 get esp from the registers (exception frame), if 0
 *		 get esp from kdba_getregcontents.
 */

static int
kdba_bt_stack_mips(struct pt_regs *regs, kdb_machreg_t *addr, int argcount,
		   struct task_struct *p, int regs_esp)
{
	kdb_machreg_t	epc, esp=0, sp_size, ra_sp;
	kdb_machreg_t	lepc, lsp;
	short		lsp_size=0;
	kdb_symtab_t	symtab;
	mm_segment_t	orgfs;

	orgfs = get_fs();
	set_fs(KERNEL_DS);

#if DEBUG_KDBA_BT
kdb_printf("kdba_bt_stack_mips(regs=0x%x)\n", (unsigned int) regs);
#endif

	kdb_printf("    ESP       EPC         Function(args)\n");

	if (addr) {
		esp = *addr ;
	} else {
		if (regs_esp)
			esp = regs_esp;
	} 
#if DEBUG_KDBA_BT
kdb_printf("kdba_bt_stack_mips() regs->cp0_epc=0x%x\n",(unsigned int) regs->cp0_epc);
#endif
	kdba_getregcontents("cp0_epc", regs, &epc);
	kdba_getregcontents("sp", regs, &esp);

	for ( lepc = epc, lsp=esp ; kdba_find_ef( lepc , &sp_size, &ra_sp); )
	{
		kdbnearsym(lepc, &symtab);
		lsp_size = sp_size ;
		if (lsp_size < 0)
		{
			lsp_size = ~lsp_size + 1;
		} else {
			kdb_printf("Error: bad sp_size %d\n",lsp_size);
			break;
		}
		if (ra_sp > lsp_size)
		{
			kdb_printf("Error: ra_sp(%d) > lsp_size(%d)\n",ra_sp,
					(int) lsp_size);
			break;
		}

		bt_print_one(lepc, lsp, lsp_size, &symtab, 0);
		lepc = *(unsigned int*) (lsp + ra_sp);

		if ( lepc & 0x3) /* lepc not aligned, must be bad */
			break;

		lsp += lsp_size;
	}

	set_fs(orgfs);

	return 0;
}

/*
 * kdba_bt_stack
 *
 *	This function implements the 'bt' command.  Print a stack
 *	backtrace.
 *
 *	bt [<address-expression>]   (addr-exp is for alternate stacks)
 *	btp <pid>		     (Kernel stack for <pid>)
 *
 * 	address expression refers to a return address on the stack.  It
 *	may be preceeded by a frame pointer.
 *
 * Inputs:
 *	regs	registers at time kdb was entered.
 *	addr	Pointer to Address provided to 'bt' command, if any.
 *	argcount
 *	p	Pointer to task for 'btp' command.
 * Outputs:
 *	None.
 * Returns:
 *	zero for success, a kdb diagnostic if error
 * Locking:
 *	none.
 * Remarks:
 *	mds comes in handy when examining the stack to do a manual
 *	traceback.
 */

int
kdba_bt_stack(struct pt_regs *regs, kdb_machreg_t *addr, int argcount,
	      struct task_struct *p)
{
	return(kdba_bt_stack_mips(regs, addr, argcount, p, 0));
}

int
kdba_bt_process(struct task_struct *p, int argcount)
{
	struct pt_regs taskregs;

#if DEBUG_KDBA_BT
kdb_printf("kdba_bt_process( p=0x%x): &taskregs=0x%x\n",
		(unsigned int) p, (unsigned int) &taskregs);
	memset(&taskregs, 0, sizeof(taskregs));
#endif

	taskregs.regs[16] = p->thread.reg16;
	taskregs.regs[17] = p->thread.reg17;
	taskregs.regs[18] = p->thread.reg18;
	taskregs.regs[19] = p->thread.reg19;
	taskregs.regs[20] = p->thread.reg20;
	taskregs.regs[21] = p->thread.reg21;
	taskregs.regs[22] = p->thread.reg22;
	taskregs.regs[23] = p->thread.reg23;
	taskregs.regs[29] = p->thread.reg29;
	taskregs.regs[30] = p->thread.reg30;
	taskregs.regs[31] = p->thread.reg31;

	taskregs.cp0_epc = thread_saved_pc(&p->thread);

#if DEBUG_KDBA_BT
kdb_printf("kdba_bt_process() p->cp0_epc=0x%x\n",
		(unsigned int) taskregs.cp0_epc);
#endif

	if ( taskregs.cp0_epc < KSEG1 || taskregs.cp0_epc >= KSEG3 )
		return 0;
#if 0
	/*
	 * Since we don't really use the TSS
	 * to store register between task switches,
	 * attempt to locate real ebp (should be
	 * top of stack if task is in schedule)
	 */
	taskregs.ebp = *(kdb_machreg_t *)(taskregs.esp);

	taskregs.xcs  = __KERNEL_CS;	/* have to assume kernel space */

	if (taskregs.esp < (unsigned long)p ||
	    taskregs.esp >= (unsigned long)p + 8192) {
		kdb_printf("Stack is not in task_struct, backtrace not available\n");
		return(0);
	}
#endif

	return kdba_bt_stack_mips(&taskregs, NULL, argcount, p, 1);
}
